from .definitions import CompilerConfig, CompilerConfig
from .runner      import autodetect
from .cli         import cli_config, ask, tell

def gcc_clang(cmd, dbg):

    def cc_once(cfg):
        lib_flags     = ['-l' + lib for lib in cfg.libs]
        include_flags = ['-I' + inc for inc in cfg.includes]
        return {
            'compileflags' : (cfg.flags + include_flags + cfg.compileflags),
            'linkflags'    : (cfg.flags + include_flags + cfg.linkflags + lib_flags),
        }
    
    def cc_compile(cfg, source):
        path = cfg.targetpath(source) + '.o'
        cfg.run([
            cmd, 
            *cfg.once['compileflags'], 
            '-c', source,
            '-o', path
        ])
        return path

    def cc_link(cfg, main, objects):
        path = cfg.targetpath(main)
        cfg.run([
            cmd,
            main,
            *objects,
            *cfg.once['linkflags'],
            '-o', path,
        ])
        return path

    def cc_run(cfg, exe):
        cfg.run([exe, *cfg.options.programargs])

    def cc_debug(cfg, exe):
        cfg.run([dbg, exe])

    return CompilerConfig(
        name         = 'gcc_clang',
        compile      = cc_compile,
        link         = cc_link,
        run          = cc_run,
        debug        = cc_debug,
        once         = cc_once,
        ccsource     = f'gcc_clang(cmd=\'{cmd}\', dbg=\'{dbg}\')',
        flags        = ['-Wall'],
        debugflags   = ['-g'],
        releaseflags = ['-O'],
    )

@cli_config('gcc_clang')
def config_gcc_clang():
    return gcc_clang(
        ask((
            'Please select a compiler command', 
            '',
        ), default=autodetect(['cc', 'gcc', 'clang'])),
        ask((
            'Please select a debug command', 
            ''
        ), default=autodetect(['gdb', 'lldb']))
    )


from os.path import join

import mekpie.messages as messages

from .definitions   import Config, CompilerConfig
from .structure     import get_main_path, get_mekpy_path, get_target_path
from .cc_gcc_clang  import gcc_clang
from .cc_avr_gcc    import avr_gcc
from .cc_emscripten import emscripten
from .cli           import panic
from .util          import (
     tab,
     check_is_file,
     check_is_dir, 
     type_name, 
     exec_str,
     file_as_str,
     smkdir,
)

def get_config(options):
    return config_from_str(options, file_as_str(get_mekpy_path()))

def config_from_str(options, source):
    return config_from_dict(exec_str(source, 'mek.py', {
        'options'        : options,
        'gcc_clang'      : gcc_clang,
        'avr_gcc'        : avr_gcc,
        'emscripten'     : emscripten,
        'CompilerConfig' : CompilerConfig,
    }))

def config_from_dict(config_dict):
    return check_config(Config(**{ key : value
        for key, value
        in config_dict.items()
        if key in Config.keys
    }))

def check_config(config):
    smkdir(get_target_path())
    check_name(config.name)
    check_main(config)
    check_libs(config.libs)
    check_cc(config.cc)
    check_flags(config.flags)
    check_compileflags(config.compileflags)
    check_linkflags(config.linkflags)
    return config

def check_name(name):
    check_type('name', name, str)

def check_main(config):
    check_type('main', config.main, str)
    check_is_file(get_main_path(config.main))

def check_libs(libs):
    check_type('libs', libs, list)
    for lib in libs:
        check_type('libs', lib, str)

def check_cc(cc):
    if hasattr(cc, 'name'):
        check_type('cc.name', cc.name, str)
    else:
        check_type('cc.name', None, str)
    if hasattr(cc, 'compile'):
        check_callable('cc.compile', cc.compile)
    else:
        check_callable('cc.compile', None)
    if hasattr(cc, 'link'):
        check_callable('cc.link', cc.link)
    else:
        check_callable('cc.link', None)
    if hasattr(cc, 'run'):
        check_callable('cc.run', cc.run)
    else:
        check_callable('cc.run', None)
    if hasattr(cc, 'debug'):
        check_callable('cc.debug', cc.debug)
    else:
        check_callable('cc.debug', None)
    if hasattr(cc, 'once'):
        check_callable('cc.once', cc.once)
    else:
        check_callable('cc.once', None)

def check_flags(flags):
    check_type('flags', flags, list)
    for flag in flags:
        check_type('flags', flag, str)

def check_compileflags(compileflags):
    check_type('compilflags', compileflags, list)
    for flag in compileflags:
        check_type('compileflags', flag, str)

def check_linkflags(linkflags):
    check_type('linkflags', linkflags, list)
    for flag in linkflags:
        check_type('linkflags', flag, str)

def check_type(name, value, *expected_types):
    if all([type(value) != exp for exp in expected_types]):
        panic(messages.type_error.format(
            name,
            ' or '.join([exp.__name__ for exp in expected_types]),
            type_name(value),
            tab(get_description(name))
        ))

def check_callable(name, value):
    if not callable(value):
        panic(messages.type_error.format(
            name,
            'callable',
            type_name(value),
            tab(get_description(name))
        ))

def get_description(name):
    return {
        'name'                   : '`name` specifies the name of the project',
        'main'                   : '`main` specifies the .c file continaing `main`',
        'libs'                   : '`libs` speicfies and libraries to load',
        'cc'                     : '`cc` specifies the c compiler configuration to use',
        'cc.name'                : '`cc.name` specifies the name of the compiler configuration',
        'cc.compile'             : '`cc.compile` specifies the compilation function',
        'cc.link'                : '`cc.link` specifies the linking function',
        'cc.run'                 : '`cc.run` specifies the run function',
        'cc.debug'               : '`cc.debug` specifies the debug function',
        'cc.once'                : '`cc.once` is run once per compiler configuration',
        'cmd'                    : '`cmd` specified the c compiler command',
        'dbg'                    : '`dbg` specifies the debugger to use',
        'flags'                  : '`flags` specifies additional compiler and linker flags',
        'compileflags'           : '`compileflags` specifies additional compiler flags',
        'linkflags'              : '`linkflags` specifies additional compiler flags',
    }[name]


from inspect import stack
import shelve

from .structure import get_cache_path

def quick_cache(value=None):
    frame = stack()[1]
    identifier = f'{frame.filename}-{frame.function}'
    with shelve.open(get_cache_path()) as cache:
        if value is None:
            return cache[identifier] if identifier in cache else value
        else:
            cache[identifier] = value
            return value


from os import name

import shelve

from .definitions import CompilerConfig, CompilerConfig
from .runner      import autodetect
from .cli         import cli_config, ask, tell, panic
from .util        import list_files
from .structure   import get_src_path
from .cache       import quick_cache

MAIN = '''
#include <avr/io.h>
#include <util/delay.h>

int main(void) {
    // !!! THIS WILL ONLY WORK ON SOME DEVICES !!!
    DDRB  = 0xFF;
    PORTB = 0xFF;
    for (;;) {
        PORTB = ~0xFF;  
        _delay_ms(1000);
        PORTB = 0xFF;
        _delay_ms(1000);
    }
    return 0;
}
'''.strip()

hardware_list = [
    'avr1',
    'at90s1200',
    'attiny11',
    'attiny12',
    'avr1',
    'attiny15',
    'attiny28',
    'avr2',
    'at90s2313',
    'at90s2323',
    'at90s2333',
    'at90s2343',
    'attiny22',
    'attiny26',
    'at90s4414',
    'at90s4433',
    'at90s4434',
    'at90s8515',
    'at90c8534',
    'at90s8535',
    'avr25',
    'at86rf401',
    'ata6289',
    'ata5272',
    'ata6616c',
    'attiny13',
    'attiny13a',
    'attiny2313',
    'attiny2313a',
    'attiny24',
    'attiny24a',
    'attiny25',
    'attiny261',
    'attiny261a',
    'attiny4313',
    'attiny43u',
    'attiny44',
    'attiny44a',
    'attiny441',
    'attiny45',
    'attiny461',
    'attiny461a',
    'attiny48',
    'attiny828',
    'attiny84',
    'attiny84a',
    'attiny841',
    'attiny85',
    'attiny861',
    'attiny861a',
    'attiny87',
    'attiny88',
    'avr3',
    'atmega603',
    'at43usb355',
    'avr31',
    'atmega103',
    'at43usb320',
    'avr35',
    'at90usb82',
    'at90usb162',
    'ata5505',
    'ata6617c',
    'ata664251',
    'atmega8u2',
    'atmega16u2',
    'atmega32u2',
    'attiny167',
    'attiny1634',
    'at76c711',
    'avr4',
    'ata6285',
    'ata6286',
    'ata6612c',
    'atmega48',
    'atmega48a',
    'atmega48pa',
    'atmega48p',
    'atmega8',
    'atmega8a',
    'atmega8515',
    'atmega8535',
    'atmega88',
    'atmega88a',
    'atmega88p',
    'atmega88pa',
    'atmega8hva',
    'at90pwm1',
    'at90pwm2',
    'at90pwm2b',
    'at90pwm3',
    'at90pwm3b',
    'at90pwm81',
    'at90can32',
    'avr5',
    'at90can64',
    'at90pwm161',
    'at90pwm216',
    'at90pwm316',
    'at90scr100',
    'at90usb646',
    'at90usb647',
    'at94k',
    'atmega16',
    'ata5790',
    'ata5702m322',
    'ata5782',
    'ata6613c',
    'ata6614q',
    'ata5790n',
    'ata5795',
    'ata5831',
    'atmega161',
    'atmega162',
    'atmega163',
    'atmega164a',
    'atmega164p',
    'atmega164pa',
    'atmega165',
    'atmega165a',
    'atmega165p',
    'atmega165pa',
    'atmega168',
    'atmega168a',
    'atmega168p',
    'atmega168pa',
    'atmega169',
    'atmega169a',
    'atmega169p',
    'atmega169pa',
    'atmega16a',
    'atmega16hva',
    'atmega16hva2',
    'atmega16hvb',
    'atmega16hvbrevb',
    'atmega16m1',
    'atmega16u4',
    'atmega32',
    'atmega32a',
    'atmega323',
    'atmega324a',
    'atmega324p',
    'atmega324pa',
    'atmega325',
    'atmega325a',
    'atmega325p',
    'atmega325pa',
    'atmega3250',
    'atmega3250a',
    'atmega3250p',
    'atmega3250pa',
    'atmega328',
    'atmega328p',
    'atmega329',
    'atmega329a',
    'atmega329p',
    'atmega329pa',
    'atmega3290',
    'atmega3290a',
    'atmega3290p',
    'atmega3290pa',
    'atmega32c1',
    'atmega32hvb',
    'atmega32hvbrevb',
    'atmega32m1',
    'atmega32u4',
    'atmega32u6',
    'atmega406',
    'atmega64rfr2',
    'atmega644rfr2',
    'atmega64',
    'atmega64a',
    'atmega640',
    'atmega644',
    'atmega644a',
    'atmega644p',
    'atmega644pa',
    'atmega645',
    'atmega645a',
    'atmega645p',
    'atmega6450',
    'atmega6450a',
    'atmega6450p',
    'atmega649',
    'atmega649a',
    'atmega6490',
    'atmega6490a',
    'atmega6490p',
    'atmega649p',
    'atmega64c1',
    'atmega64hve',
    'atmega64hve2',
    'atmega64m1',
    'm3000',
    'avr51',
    'at90can128',
    'at90usb1286',
    'at90usb1287',
    'atmega128',
    'atmega128a',
    'atmega1280',
    'atmega1281',
    'atmega1284',
    'atmega1284p',
    'atmega128rfr2',
    'atmega1284rfr2',
    'avr6',
    'atmega2560',
    'atmega2561',
    'atmega256rfr2',
    'atmega2564rfr2',
    'avrxmega2',
    'atxmega16a4',
    'atxmega16a4u',
    'atxmega16c4',
    'atxmega16d4',
    'atxmega32a4',
    'atxmega32a4u',
    'atxmega32c3',
    'atxmega32c4',
    'atxmega32d3',
    'atxmega32d4',
    'atxmega8e5',
    'atxmega16e5',
    'atxmega32e5',
    'avrxmega4',
    'atxmega64a3',
    'atxmega64a3u',
    'atxmega64a4u',
    'atxmega64b1',
    'atxmega64b3',
    'atxmega64c3',
    'atxmega64d3',
    'atxmega64d4',
    'avrxmega5',
    'atxmega64a1',
    'atxmega64a1u',
    'avrxmega6',
    'atxmega128a3',
    'atxmega128a3u',
    'atxmega128b1',
    'atxmega128b3',
    'atxmega128c3',
    'atxmega128d3',
    'atxmega128d4',
    'atxmega192a3',
    'atxmega192a3u',
    'atxmega192c3',
    'atxmega192d3',
    'atxmega256a3',
    'atxmega256a3u',
    'atxmega256a3b',
    'atxmega256a3bu',
    'atxmega256c3',
    'atxmega256d3',
    'atxmega384c3',
    'atxmega384d3',
    'atxmega128a1',
    'avrxmega7',
    'atxmega128a1u',
    'atxmega128a4u',
    'avrtiny10',
    'attiny4',
    'attiny5',
    'attiny9',
    'attiny20',
    'attiny40',
]

programmer_list = [
    'arduino',
    'avr910',
    'avrftdi',
    'buspirate',
    'buspirate_bb',
    'butterfly',
    'butterfly_mk',
    'dragon_dw',
    'dragon_hvsp',
    'dragon_isp',
    'dragon_jtag',
    'dragon_pdi',
    'dragon_pp',
    'ftdi_syncbb',
    'jtagmki',
    'jtagmkii',
    'jtagmkii_avr32',
    'jtagmkii_dw',
    'jtagmkii_isp',
    'jtagmkii_pdi',
    'jtagice3',
    'jtagice3_pdi',
    'jtagice3_dw',
    'jtagice3_isp',
    'linuxgpio',
    'par',
    'pickit2',
    'serbb',
    'stk500',
    'stk500generic',
    'stk500v2',
    'stk500hvsp',
    'stk500pp',
    'stk600',
    'stk600hvsp',
    'stk600pp',
    'usbasp',
    'usbtiny',
    'wiring',
]

def avr_gcc(hardware, programmer, baud):

    def cc_once(cfg):
        lib_flags     = ['-l' + lib for lib in cfg.libs]
        include_flags = ['-I' + inc for inc in cfg.includes]
        hardware_flag = [f'-mmcu={hardware}']
        return {
            'compileflags' : (cfg.flags + include_flags + cfg.compileflags + hardware_flag),
            'linkflags'    : (cfg.flags + include_flags + cfg.linkflags + lib_flags + hardware_flag),
        }
    
    def cc_compile(cfg, source):
        path = cfg.targetpath(source) + '.o'
        cfg.run([
            'avr-gcc', 
            *cfg.once['compileflags'], 
            '-c', source,
            '-o', path
        ])
        return path

    def cc_link(cfg, main, objects):
        path_elf = cfg.targetpath(main) + '.elf'
        path_hex = cfg.targetpath(main) + '.hex' 
        asm = list_files(get_src_path(), with_ext='S')
        cfg.run([
            'avr-gcc',
            main,
            *objects,
            *asm,
            *cfg.once['linkflags'],
            '-o', path_elf,
        ])
        cfg.run([
            'avr-objcopy',
            '-O', 'ihex',
            path_elf,
            path_hex,
        ])
        return path_hex

    def cc_run(cfg, exe):

        @cli_config('port')
        def config_port():
            port = quick_cache()
            if name == 'posix':
                tell('Available ports:')
                cfg.run(['ls /dev/cu.*'], shell=True)
            return quick_cache(ask(('Please enter your device port', ''), default=port))

        cfg.run([
            'avrdude',
            '-v',
            '-p', hardware,
            '-c', programmer,
            '-P', config_port(),
            '-b', baud,
            '-D',
            '-U', f'flash:w:{exe}:i',
        ])
        input('Press <enter> to continue...')

    def cc_debug(cfg, exe):
        panic('Debug is not supported for avr_gcc compiler config!')

    return CompilerConfig(
        name         = 'avr_gcc',
        compile      = cc_compile,
        link         = cc_link,
        run          = cc_run,
        debug        = cc_debug,
        once         = cc_once,
        ccsource     = f'avr_gcc(hardware=\'{hardware}\', programmer=\'{programmer}\', baud=\'{baud}\')',
        csource      = MAIN,
        flags        = ['-Wall'],
        debugflags   = ['-g'],
        releaseflags = ['-O'],
    )

@cli_config('avr_gcc')
def config_avr_gcc():
    return avr_gcc(
        ask((
            'Please select your AVR hardware', 
            '`{}` is not a valid AVR hardware selection!',
        ), options=hardware_list),
        ask((
            'Please select your programmer', 
            '`{}` is not a valid programmer',
        ), default='wiring', options=programmer_list),
        ask((
            'Please select your baud rate',
            '',        
        ), default='115200')
    )


from .definitions import VERSION

usage = '''
Usage:
    mekpie [options] `command` [command args...] -- [program args...]

Options:
    -h, --help      Display this message
    -V, --version   Print version info and exit
    -q, --quiet     No output to stdout
    -r, --release   Run the command for release
    -m, --mode      Provide a mode
    -c, --changedir Run the mekpie command in the provided directory
    -d, --developer Run mekpie in developer mode

Commands:
    `new`   name     Create a new mekpie project
    `test`  names... Build and execute tests or test
    `clean`          Remove artefacts from previous builds
    `build`          Compile the current project
    `run`            Build and execute main
    `debug`          Build and execute under a debugger
    `dist`           Create an executable in the project root

'''

version = f'''
    mekpie version {VERSION}
'''

no_command = '''
!! Invalid command: No command provided!

    mekpie expects a command following its invocation.
    The following commands are available:

        `help`    Display a help message
        `version` Print version info and exit
        `new`     Create a new mekpie project
        `init`    Create a project in an existing directory
        `test`    Build and execute tests or test
        `clean`   Remove artefacts from previous builds
        `build`   Compile the current project
        `run`     Build and execute main
        `debug`   Build and execute under a debugger
        `dist`    Create an executable in the project root
'''

repeated_option = '!! Repeated option: option {} appears twice!'

too_many_arguments = '!! Unexpected argument: Too many arguments provided!'

unknown_argument = '!! Unexpected argument: Unknown argument provided!'

name_cannot_already_exist = '''
!! Invalid Argument: Cannot create a project in a directory that already exists!

    The directory "./{}" already exists. If you want to use this folder with
    mekpie, navigate within the folder, and run the `init` command.
'''

name_cannot_be_empty = '''
!! Missing argument: You must provide the positional argument `name`!

    When creating a project with `new` or `init` you must provide a positional
    argument `name` after the command.
'''

execution_error = '''
!! Execution error: Error while trying to execute a python resource!
    
    Error occured when attempting to execute: {}

{}
'''

compiler_config_error = '''
!! Config error: Invalid compiler configuration!

    Exptected one of the following: {}
    Instead found a value of: {}
'''

type_error = '''
!! Config error: Invalid type!

    Expected `{}` to be of type: {}
    Instead found a value of type: {}
{}
'''

file_not_found = '''
!! File not found: Could not find "{}"!
'''

directory_not_found = '''
!! Directory not found: Could not find "{}"!
'''

compiler_flag_error = '''
!! Flags not found: Could not find compiler flag for "{}" with purpose "{}"!

    Add the following entry to mek.py to include the correct flag:
    {}
'''

no_tests = '''
!! No tests found: Could not find any tests to run in /tests/!

    Try adding a c source file with a `main` function to /tests/
'''

no_tests_with_name = '''
!! No tests found: Could not find any test with the name {} in /tests/!

     Check that you spelled the test name correctly.
'''

api_no_options = '''
!! No options: API call to mekpie was made without an `Options` instance!

     Call `core.mekpie` with an instance of `definitions.Options`
'''

failed_compiler_call = '''
!! Build failed: the following call to the compiler failed:

    {}

And the following errors were produced:
    {}
'''

failed_program_call = '''
!! Call failed: the following program call failed:

    {}
'''

release_debug = '''
!! Build failed: cannot launch debugger for a release build

    To launch the debugger run the command again without the --release flag.
'''

created = '''
{} created successfully!
'''

initialized = '''
{} initialized successfully!
'''

clean = '''
Project successfully cleaned.
'''

build_succeeded = '''
Project successfully built. ({:.3f}s)
'''

mekpie_config_name = (
    'Please provide a name for your project',
    '{} is not a valid project name!'
)

compiler_configs = '''
Mekpie supports the following c compilers, please select one (compilers must be installed seperately):
    - `gcc_clang`   for use with the gcc or clang compiler 
    - `emscripten`  for use with the emscripten c to js compiler
    - `avr_gcc`     for use with avr-gcc and avrdude
'''

mekpie_config_cc = (
    'Please select a cc',
    '{} is not a valid cc! See the list above.'
)


from time               import time
from os.path            import join
from collections        import namedtuple
from concurrent.futures import ThreadPoolExecutor

import mekpie.messages as messages

from .runner    import lrun
from .cli       import panic, tell
from .util      import (
    smv,
    clamp,
    first,
    rest,
    list_files, 
    list_all_dirs,
    remove_contents, 
    filename, 
)
from .structure import (
    get_project_path,
    get_test_path,
    get_target_path,
    get_target_debug_path,
    get_target_release_path,
    get_main_path,
    get_src_path,
    get_includes_path,
)

max_threads = 32

def command_clean(cfg):
    remove_contents(get_target_debug_path())
    remove_contents(get_target_release_path())
    if not cfg.options.quiet:
        tell(messages.clean.strip())

def command_build(cfg):
    command_clean(cfg)
    start   = time()
    cfg.includes   = get_includes_paths()
    cfg.targetpath = lambda path : get_bin_path(cfg, path)
    cfg.once       = cfg.cc.once(cfg)
    
    objects = compile_objects(cfg)
    exes    = link_exes(cfg, objects)

    if not cfg.options.quiet:
        tell(messages.build_succeeded.format(time() - start).strip())

    return exes

def compile_objects(cfg):
    sources = get_sources(cfg)
    runset  = []
    cfg.run = lambda args, **kwargs : lrun(args, 
        quiet  = cfg.options.quiet, 
        error  = False, 
        runset = runset,
        **kwargs,
    ) 
    with ThreadPoolExecutor(max_workers=clamp(len(sources), 1, max_threads)) as e:
        objects = e.map(lambda source: cfg.cc.compile(cfg, source), sources)
    if not all(runset):
        exit(1)
    return list(objects)

def link_exes(cfg, objects):
    tests = (list_files(get_test_path(), with_ext='.c') 
        + list_files(get_test_path(), with_ext='.cpp') 
        + list_files(get_test_path(), with_ext='.cc')
    )
    mains   = [get_main_path(cfg.main), *tests]
    runset  = []
    cfg.run = lambda args, **kwargs : lrun(args, 
        quiet  = True, 
        error  = False, 
        runset = runset,
        **kwargs
    ) 
    with ThreadPoolExecutor(max_workers=clamp(len(mains), 1, max_threads)) as e:
        exes = e.map(lambda main: cfg.cc.link(cfg, main, objects), mains)
    if not all(runset):
        cfg.run = lambda args: lrun(args, cfg.options.quiet)
        for main in mains:
            cfg.cc.link(cfg, main, objects)
    return list(exes)

def command_run(cfg):
    exes = command_build(cfg)
    exe  = first(exes)
    cfg.run = lambda args, **kwargs: lrun(args, cfg.options.quiet, **kwargs)
    cfg.cc.run(cfg, exe)

def command_debug(cfg):
    if cfg.options.release:
        panic(messages.release_debug)
    exes = command_build(cfg)
    exe  = first(exes)
    cfg.run = lambda args, **kwargs: lrun(args, cfg.options.quiet, **kwargs)
    cfg.cc.debug(cfg, exe)

def command_test(cfg):
    exes  = command_build(cfg)
    tests = rest(exes) 
    cfg.run = lambda args, **kwargs: lrun(args, cfg.options.quiet, **kwargs)
    for test in tests:
        if len(cfg.options.commandargs) == 0 or any(name in test for name in cfg.options.commandargs):
            cfg.cc.run(cfg, test)

def command_dist(cfg):
    exes = command_build(cfg)
    exe  = first(exes)
    smv(exe, join(get_project_path(), cfg.name))

def get_bin_path(cfg, path):
    root = get_target_release_path() if cfg.options.release else get_target_debug_path()
    return join(root, filename(path))

def get_sources(cfg):
    sources = (list_files(get_src_path(), with_ext='.c') 
        + list_files(get_src_path(), with_ext='.cpp') 
        + list_files(get_src_path(), with_ext='.cc')
    )
    sources.remove(get_main_path(cfg.main))
    return sources

def get_includes_paths():
    includes = get_includes_path()
    return [includes] + list_all_dirs(includes)